package de.unisaar.jfap.ex7;

import de.unisaar.jfap.ex7.visitors.DfsGraphVisitor;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Visitor collecting the discovery and finishing times of the vertices and classifying the edges
 * into tree, back, forward and cross edges.
 *
 * @author Bernd Kiefer, DFKI
 * @author Jörg Steffen, DFKI
 * @param <E> the type of information assigned to the edges of the traversed graph
 * @param <V> the type of information assigned to the vertices of the traversed graph
 */
public class CollectTimesVisitor<E, V> implements DfsGraphVisitor<E, V> {

  private static final Logger logger = LoggerFactory.getLogger(CollectTimesVisitor.class);

  // the discovery times of the vertices
  private Map<Integer, Integer> discovery;

  // the finishing times of the vertices
  private Map<Integer, Integer> finish;

  private List<Edge<E>> treeEdges;
  private List<Edge<E>> backEdges;
  private List<Edge<E>> forwardEdges;
  private List<Edge<E>> crossEdges;

  // keeps track of the time counter
  private int time;

  /** Create a new time collector visitor. */
  public CollectTimesVisitor() {
    this.time = 1;
    this.discovery = new HashMap<>();
    this.finish = new HashMap<>();
    this.treeEdges = new ArrayList<>();
    this.backEdges = new ArrayList<>();
    this.forwardEdges = new ArrayList<>();
    this.crossEdges = new ArrayList<>();
  }

  public Map<Integer, Integer> getDiscoveryMap() {
    return this.discovery;
  }

  public Map<Integer, Integer> getFinishMap() {
    return this.finish;
  }

  public List<Edge<E>> getTreeEdges() {
    return this.treeEdges;
  }

  public List<Edge<E>> getBackEdges() {
    return this.backEdges;
  }

  public List<Edge<E>> getForwardEdges() {
    return this.forwardEdges;
  }

  public List<Edge<E>> getCrossEdges() {
    return this.crossEdges;
  }

  @Override
  public void startVertex(int v, DirectedGraph<E, V> g) {
    logger.info("starting DFS at vertex {}", g.getVertexProperty(v));
  }

  @Override
  public void discoverVertex(int v, DirectedGraph<E, V> g) {
    this.discovery.put(v, this.time);
    this.time++;
  }

  @Override
  public void finishVertex(int v, DirectedGraph<E, V> g) {
    this.finish.put(v, this.time);
    this.time++;
  }

  @Override
  public void treeEdgeBefore(Edge<E> e, DirectedGraph<E, V> g) {
    this.treeEdges.add(e);
  }

  @Override
  public void treeEdgeAfter(Edge<E> e, DirectedGraph<E, V> g) {
    // nothing to do here
  }

  @Override
  public void nonTreeEdge(Edge<E> e, DirectedGraph<E, V> g) {
    int startDiscovery = this.discovery.getOrDefault(e.getStartVertex(), 0);
    int endDiscovery = this.discovery.getOrDefault(e.getEndVertex(), 0);
    int endFinish = this.finish.getOrDefault(e.getEndVertex(), 0);
    if (endFinish == 0 && startDiscovery > endDiscovery) {
      this.backEdges.add(e);
    } else if (endFinish != 0 && endDiscovery < startDiscovery) {
      this.crossEdges.add(e);
    } else if (endFinish != 0 && endDiscovery > startDiscovery) {
      this.forwardEdges.add(e);
    } else {
      logger.error("edge {} not classified", e);
    }
  }
}
