package de.unisaar.jfap.ex7;

import de.unisaar.jfap.ex7.BinaryPredicate.EqualsPredicate;
import de.unisaar.jfap.ex7.visitors.BfsGraphVisitor;
import de.unisaar.jfap.ex7.visitors.DfsGraphVisitor;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Directed graph with vertices represented as ints, their edges and assigned properties are managed
 * in maps using the ints as keys.
 *
 * @author Bernd Kiefer, DFKI
 * @author Jörg Steffen, DFKI
 * @param <E> edge info used in this graph
 * @param <V> vertex properties used in this graph
 */
public class DirectedGraph<E, V> {

  private static final Logger logger = LoggerFactory.getLogger(DirectedGraph.class);

  private AtomicInteger idGenerator = new AtomicInteger();

  // edges
  private Map<Integer, List<Edge<E>>> outEdges;
  private Map<Integer, List<Edge<E>>> inEdges;

  // info assigned to vertices
  private Map<Integer, V> vertexPropertyMap;

  /** Create a new directed graph. */
  public DirectedGraph() {
    this.outEdges = new LinkedHashMap<>();
    this.inEdges = new LinkedHashMap<>();
    this.vertexPropertyMap = new LinkedHashMap<>();
  }

  /**
   * Return all vertices of this graph.
   *
   * @return all vertices
   */
  public Collection<Integer> getVertices() {
    return this.outEdges.keySet();
  }

  /**
   * Return the number of vertices in this graph.
   *
   * @return number of vertices
   */
  public int getNumberOfVertices() {
    return this.outEdges.size();
  }

  /**
   * Check if the given vertex is a vertex of this graph.
   *
   * @param vertex the vertex to check
   * @return {@code true} if the vertex is non-deleted, {@code false} otherwise
   */
  public boolean isVertex(int vertex) {
    return this.outEdges.containsKey(vertex);
  }

  /**
   * Return the edges emerging from the given vertex.
   *
   * @param vertex the vertex
   * @return the outgoing edges; may be empty
   */
  public List<Edge<E>> getOutEdges(int vertex) {
    List<Edge<E>> result = this.outEdges.get(vertex);
    return (result == null) ? Collections.emptyList() : Collections.unmodifiableList(result);
  }

  /**
   * Return the edges ending in the given vertex.
   *
   * @param vertex the vertex
   * @return the incoming edges; may be empty
   */
  public List<Edge<E>> getInEdges(int vertex) {
    List<Edge<E>> result = this.inEdges.get(vertex);
    return (result == null) ? Collections.emptyList() : Collections.unmodifiableList(result);
  }

  /**
   * Get property stored with the given vertex.
   *
   * @param vertex the vertex
   * @return the property assigned to the vertex, {@code null} if not set
   */
  public V getVertexProperty(int vertex) {
    return this.vertexPropertyMap.get(vertex);
  }

  /**
   * Get map of vertices to their properties.
   *
   * @return vertex-to-property map
   */
  public Map<Integer, V> getVertexPropertyMap() {
    return this.vertexPropertyMap;
  }

  /**
   * Create a new vertex without any assigned property. Also initialize edges lists.
   *
   * @return the new vertex
   */
  public int newVertex() {
    int vertex = idGenerator.getAndIncrement();
    this.outEdges.put(vertex, new ArrayList<>(3));
    this.inEdges.put(vertex, new ArrayList<>(3));
    return vertex;
  }

  /**
   * Create a new vertex with the given assigned property. Also initialize edges lists.
   *
   * @param value property to assign with the new vertex
   * @return the new vertex
   */
  public int newVertex(V value) {
    int vertex = newVertex();
    this.vertexPropertyMap.put(vertex, value);
    return vertex;
  }

  /**
   * Create a new edge with the given edge info and returns it.
   *
   * @param edgeInfo the info assigned to the newly created edge
   * @param from the start vertex of the edge, which must be a valid vertex of the graph
   * @param to the end vertex of the edge, which must be a valid vertex of the graph
   * @return a newly created edge or {@code null} if a non-valid vertex is given
   */
  public Edge<E> newEdge(E edgeInfo, int from, int to) {
    if (isVertex(from) && isVertex(to)) {
      Edge<E> edge = new Edge<>(edgeInfo);
      // update from
      edge.setStartVertex(from);
      List<Edge<E>> outEdgeList = this.outEdges.get(from);
      outEdgeList.add(edge);
      // update to
      edge.setEndVertex(to);
      List<Edge<E>> inEdgeList = this.inEdges.get(to);
      inEdgeList.add(edge);
      return edge;
    }
    return null;
  }

  /**
   * Remove the given edge by modifying the edge lists of the adjacent vertices.
   *
   * @param edge the edge to remove
   */
  public void removeEdge(Edge<E> edge) {
    // outgoing edges of start vertex
    List<Edge<E>> startVertexOut = this.outEdges.get(edge.getStartVertex());
    startVertexOut.remove(edge);
    // incoming edges of end vertex
    List<Edge<E>> endVertexIn = this.inEdges.get(edge.getEndVertex());
    endVertexIn.remove(edge);
  }

  /**
   * Remove the given vertex from this directed graph.
   *
   * <p>Requires:
   *
   * <ul>
   *   <li>removing all edges in and out of the vertex
   *   <li>removing all properties associated with the vertex
   * </ul>
   *
   * @param vertex the vertex to remove from the graph
   */
  public void removeVertex(int vertex) {

    if (!isVertex(vertex)) {
      // nothing to do
      return;
    }
    // remove the edges from and to this vertex from other vertices edge lists

    // outgoing edges:
    List<Edge<E>> out = this.outEdges.get(vertex);
    for (Edge<E> edge : out) {
      // remove edge from incoming edges lists of end vertex
      List<Edge<E>> endVertexIn = this.inEdges.get(edge.getEndVertex());
      endVertexIn.remove(edge);
    }
    this.outEdges.remove(vertex);

    // incoming edges:
    List<Edge<E>> in = this.inEdges.get(vertex);
    for (Edge<E> edge : in) {
      // remove edge from outgoing edges lists of start vertex
      List<Edge<E>> startVertexOut = this.outEdges.get(edge.getStartVertex());
      startVertexOut.remove(edge);
    }
    this.inEdges.remove(vertex);

    // clear the entries in the property map
    this.vertexPropertyMap.remove(vertex);
  }

  /**
   * Change the start vertex of this edge to the given vertex.
   *
   * @param edge the edge for which to change the start vertex
   * @param startVertex the new start vertex
   */
  public void changeStartVertex(Edge<E> edge, int startVertex) {
    // remove edge from old start vertex's outgoing edge list
    List<Edge<E>> oldOutEdges = this.outEdges.get(edge.getStartVertex());
    oldOutEdges.remove(edge);
    // set new start vertex
    edge.setStartVertex(startVertex);
    List<Edge<E>> newOutEdges = this.outEdges.get(startVertex);
    newOutEdges.add(edge);
  }

  /**
   * Change the end vertex of this edge to the given vertex.
   *
   * @param edge the edge for which to change the end vertex
   * @param endVertex the new end vertex
   */
  public void changeEndVertex(Edge<E> edge, int endVertex) {
    // remove edge from old end vertex's incoming edge list
    List<Edge<E>> oldInEdges = this.inEdges.get(edge.getEndVertex());
    oldInEdges.remove(edge);
    // set new end vertex
    edge.setEndVertex(endVertex);
    List<Edge<E>> newInEdges = this.inEdges.get(endVertex);
    newInEdges.add(edge);
  }

  /**
   * Return the first edge emerging from the given vertex where the edge info is compatible to the
   * given edge info according to the given comparator.
   *
   * @param vertex the vertex
   * @param edgeInfo the edge info for which to find matching edges
   * @param comp a comparator to compare the edge info
   * @return the first edge that matches or {@code null} if no matching edge is found
   */
  public Edge<E> findEdge(int vertex, E edgeInfo, Comparator<E> comp) {
    List<Edge<E>> edges = this.outEdges.get(vertex);
    if (edges == null) {
      return null;
    }
    for (Edge<E> oneEdge : edges) {
      if (comp.compare(oneEdge.getInfo(), edgeInfo) == 0) {
        return oneEdge;
      }
    }
    return null;
  }

  /**
   * Return all edges emerging from the given vertex where the edge info is compatible to the given
   * edge info according to the given comparator.
   *
   * @param vertex the vertex
   * @param edgeInfo the edge info for which to find matching edges
   * @param comp a comparator to compare the edge info
   * @return the list of matching edges; may be empty
   */
  public List<Edge<E>> findEdges(int vertex, E edgeInfo, Comparator<E> comp) {
    List<Edge<E>> result = new ArrayList<>();
    List<Edge<E>> edges = this.outEdges.get(vertex);
    for (Edge<E> oneEdge : edges) {
      if (comp.compare(oneEdge.getInfo(), edgeInfo) == 0) {
        result.add(oneEdge);
      }
    }
    return result;
  }

  /**
   * Return all edges from this directed graph where the edge info is compatible to the given edge
   * info according to the given comparator.
   *
   * @param edgeInfo the edge info for which to find matching edges
   * @param comp a comparator to compare the edge info
   * @return the list of matching edges; may be empty
   */
  public List<Edge<E>> findEdges(E edgeInfo, Comparator<E> comp) {
    List<Edge<E>> result = new ArrayList<>();
    for (int vertex = 0; vertex < this.outEdges.size(); ++vertex) {
      if (this.outEdges.get(vertex) != null) {
        result.addAll(findEdges(vertex, edgeInfo, comp));
      }
    }
    return result;
  }

  /**
   * Return all vertices whose assigned property matches the given value using the given predicate.
   *
   * @param val the value to compare against
   * @param pred the predicate used to compare
   * @return a list of matching vertices
   */
  public List<Integer> findVertices(V val, BinaryPredicate<V> pred) {
    List<Integer> result = new ArrayList<>();
    for (int i : this.outEdges.keySet()) {
      if (pred.compare(this.vertexPropertyMap.get(i), val)) {
        result.add(i);
      }
    }
    return result;
  }

  /**
   * Return first vertex whose assigned property matches the given value using the given predicate.
   *
   * @param val the value to compare against
   * @param pred the predicate used to compare
   * @return the first vertex that matches or {@code -1} if no matching vertex is found
   */
  public int findVertex(V val, BinaryPredicate<V> pred) {
    for (int i : this.outEdges.keySet()) {
      if (pred.compare(this.vertexPropertyMap.get(i), val)) {
        return i;
      }
    }
    return -1;
  }

  /**
   * Return a string representation of the given vertex, optionally using the properties assigned to
   * the vertex.
   *
   * @param vertex the vertex
   * @param useProperties if {@code true} use the property assigned to the vertex, use vertex number
   *     if {@code false}
   * @return a string representation of the vertex
   */
  public String toString(int vertex, boolean useProperties) {

    StringBuilder strRep = new StringBuilder();

    if (!this.outEdges.get(vertex).isEmpty()) {
      for (Edge<E> edge : this.outEdges.get(vertex)) {
        strRep.append(String.format("%s%n", edge.toString(this.vertexPropertyMap)));
      }
    }
    // ignore incoming edges since they are already printed as outgoing edges
    // of other vertices
    if (this.outEdges.get(vertex).isEmpty() && this.inEdges.get(vertex).isEmpty()) {
      // print vertices having no edges
      if (useProperties) {
        strRep.append(this.vertexPropertyMap.get(vertex));
      } else {
        strRep.append(vertex + "");
      }
      strRep.append(String.format("%n"));
    }
    return strRep.toString();
  }

  /**
   * Return a string representation of this graph, optionally using the properties assigned to the
   * vertices.
   *
   * @param useProperties if {@code true} use the property assigned to the vertex, use vertex number
   *     if {@code false}
   * @return a string representation of this graph
   */
  public String toString(boolean useProperties) {

    StringBuilder strRep = new StringBuilder();

    for (int vertex : this.outEdges.keySet()) {
      strRep.append(toString(vertex, useProperties));
    }

    return strRep.toString();
  }

  @Override
  public String toString() {
    return toString(true);
  }

  /**
   * Recursive helper function for the dfs methods.
   *
   * <p>The U set is implemented using the given BitSet {@code visited}. The desired functionality
   * can be implemented by the given {@link DfsGraphVisitor} argument.
   *
   * @param vertex the vertex where to start the recursive procedure
   * @param visitor the DFS visitor to use
   * @param visited the implementation of the U set (see slides)
   */
  void dfsVisit(int vertex, DfsGraphVisitor<E, V> visitor, BitSet visited) {
    // vertex gets grey
    visited.set(vertex);
    visitor.discoverVertex(vertex, this);
    for (Edge<E> edge : getOutEdges(vertex)) {
      int target = edge.getEndVertex();
      if (!visited.get(target)) {
        // white target vertex, so we found a tree edge
        visitor.treeEdgeBefore(edge, this);
        dfsVisit(target, visitor, visited);
        visitor.treeEdgeAfter(edge, this);
      } else {
        // target black or grey, so we found a non-tree edge
        visitor.nonTreeEdge(edge, this);
      }
    }
    // vertex gets black
    visitor.finishVertex(vertex, this);
  }

  /**
   * Visit all vertices of this graph in a depth first manner. The vertex list is traversed in
   * order, and every non-visited vertex is then taken as start vertex, consecutively. The desired
   * functionality can be implemented by the given {@link DfsGraphVisitor} argument.
   *
   * @param visitor the DFS visitor to use
   */
  public void dfs(DfsGraphVisitor<E, V> visitor) {
    BitSet visited = new BitSet();
    for (int vertex : this.outEdges.keySet()) {
      if (!visited.get(vertex)) {
        visitor.startVertex(vertex, this);
        dfsVisit(vertex, visitor, visited);
      }
    }
  }

  /**
   * Visit the vertices of this graph that are reachable from vertex in a depth first manner. The
   * desired functionality can be implemented by the given {@link DfsGraphVisitor} argument.
   *
   * @param vertex the vertex where to start
   * @param visitor the DFS visitor to use
   */
  public void dfs(int vertex, DfsGraphVisitor<E, V> visitor) {
    visitor.startVertex(vertex, this);
    dfsVisit(vertex, visitor, new BitSet());
  }

  /**
   * Helper function for the bfs methods.
   *
   * <p>Visits all vertices of this graph in a breadth first manner. The vertex list is traversed in
   * order, and every non-visited vertex is then put onto a queue, consecutively. The desired
   * functionality can be implemented by the given {@link BfsGraphVisitor} argument.
   *
   * @param vertex the vertex where to start the procedure
   * @param visitor the BFS visitor to use
   * @param visited the implementation of the U set (see slides)
   */
  void bfsVisit(int vertex, BfsGraphVisitor<E, V> visitor, BitSet visited) {
    Queue<Integer> active = new LinkedList<>();
    active.offer(vertex);
    // vertex gets gray
    visited.set(vertex);
    visitor.discoverVertex(vertex, this);
    while (!active.isEmpty()) {
      int nextVertex = active.poll();
      for (Edge<E> outEdge : getOutEdges(nextVertex)) {
        int target = outEdge.getEndVertex();
        if (!visited.get(target)) {
          // white target vertex, so we found a tree edge
          visitor.treeEdge(outEdge, this);
          active.offer(target);
          // vertex gets gray
          visited.set(target);
          visitor.discoverVertex(target, this);
        } else {
          // grey target vertex, so we found a non-tree edge
          visitor.nonTreeEdge(outEdge, this);
        }
      }
      // vertex gets black
      visitor.finishVertex(nextVertex, this);
    }
  }

  /**
   * Visit all vertices of this graph in a breadth first manner. The vertex list is traversed in
   * order, and every non-visited vertex is then put onto a queue, consecutively. The desired
   * functionality can be implemented by the given {@link BfsGraphVisitor} argument.
   *
   * @param visitor the BFS visitor to use
   */
  public void bfs(BfsGraphVisitor<E, V> visitor) {
    BitSet visited = new BitSet();
    for (int vertex : this.outEdges.keySet()) {
      if (!visited.get(vertex)) {
        visitor.startVertex(vertex, this);
        bfsVisit(vertex, visitor, visited);
      }
    }
  }

  /**
   * Visit the vertices of this graph that are reachable from vertex in a breadth first manner. The
   * desired functionality can be implemented by the given {@link BfsGraphVisitor} argument.
   *
   * @param vertex the vertex where to start
   * @param visitor the BFS visitor to use
   */
  public void bfs(int vertex, BfsGraphVisitor<E, V> visitor) {
    visitor.startVertex(vertex, this);
    bfsVisit(vertex, visitor, new BitSet());
  }

  /**
   * Create a graph from a readable specification. Example:
   *
   * <pre>
   * A --> B D
   * B --> C D
   * C --> B
   * D --> C
   * </pre>
   *
   * @param graphString the string from which to read the graph
   * @return the graph
   */
  public static DirectedGraph<String, String> readGraph(String graphString) {
    DirectedGraph<String, String> graph = new DirectedGraph<>();
    EqualsPredicate<String> eqPred = new EqualsPredicate<>();
    for (String oneLine : graphString.lines().toList()) {
      String[] fromTo = oneLine.split("\\s*-->\\s*");
      if (fromTo.length != 2) {
        logger.error("ignoring line {}", oneLine);
        continue;
      }
      int from = graph.findVertex(fromTo[0], eqPred);
      if (from == -1) {
        // if start vertex doesn't exist, create it
        from = graph.newVertex(fromTo[0]);
      }
      String[] endLabels = fromTo[1].split("\\s+");
      for (String oneEndLabel : endLabels) {
        // add outgoing edges
        int to = graph.findVertex(oneEndLabel, eqPred);
        if (to == -1) {
          // if target vertex doesn't exist, create it
          to = graph.newVertex(oneEndLabel);
        }
        graph.newEdge(fromTo[0] + oneEndLabel, from, to);
      }
    }
    return graph;
  }

  /**
   * Create a graph from a readable specification including edge weights. Example:
   *
   * <pre>
   * s --> w(2) z(4)
   * z --> w(5)
   * w --> x(6)
   * x --> z(8) s(7)
   * </pre>
   *
   * @param graphString the string from which to read the graph
   * @return the graph
   */
  public static DirectedGraph<Integer, String> readGraphWithWeights(String graphString) {
    DirectedGraph<Integer, String> graph = new DirectedGraph<>();
    EqualsPredicate<String> eqPred = new EqualsPredicate<>();
    Pattern labelWithWeightPattern = Pattern.compile("([^(]*)\\(([0-9]*)\\)");
    for (String oneLine : graphString.lines().toList()) {
      String[] fromTo = oneLine.split("\\s*-->\\s*");
      if (fromTo.length != 2) {
        logger.error("ignoring line {}", oneLine);
        continue;
      }
      int from = graph.findVertex(fromTo[0], eqPred);
      if (from == -1) {
        // if start vertex doesn't exist, create it
        from = graph.newVertex(fromTo[0]);
      }
      String[] endLabelWithWeight = fromTo[1].split("\\s+");
      for (String oneEndLabelWithWeight : endLabelWithWeight) {
        // add outgoing edges
        Matcher matcher = labelWithWeightPattern.matcher(oneEndLabelWithWeight);
        if (!matcher.matches()) {
          logger.error("label with weight not matched in {}", oneEndLabelWithWeight);
          continue;
        }
        String label = matcher.group(1);
        int weight = Integer.parseInt(matcher.group(2));
        int to = graph.findVertex(label, eqPred);
        if (to == -1) {
          // if target vertex doesn't exist, create it
          to = graph.newVertex(label);
        }
        graph.newEdge(weight, from, to);
      }
    }
    return graph;
  }
}
